// Copyright (C) 2006  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.

#ifdef DLIB_ALL_SOURCE_END
#include "dlib_basic_cpp_build_tutorial.txt"
#endif

#ifndef DLIB_IMAGE_TRANSFORMs_
#define DLIB_IMAGE_TRANSFORMs_ 

#include "image_transforms/assign_image.h"
#include "image_transforms/equalize_histogram.h"
#include "image_transforms/morphological_operations.h"
#include "image_transforms/spatial_filtering.h"
#include "image_transforms/thresholding.h"
#include "image_transforms/edge_detector.h"
#include "image_transforms/draw.h"
#include "image_transforms/integral_image.h"
#include "image_transforms/image_pyramid.h"
#include "image_transforms/hough_transform.h"
#include "image_transforms/label_connected_blobs.h"
#include "image_transforms/colormaps.h"
#include "image_transforms/segment_image.h"
#include "image_transforms/interpolation.h"
#include "image_transforms/fhog.h"
#include "image_transforms/lbp.h"

#endif // DLIB_IMAGE_TRANSFORMs_

